##################################################################################
# ECR - January 2017
#
# These are the R functions accompanying
# E. Cantoni & M. Auda (2017)
# "Stochastic Variable Selection selection strategies for zero-inflated models"
#################################################################################


# aicMHzip function
# Funtion to generate the Markov chain, when using AIC
aicMHzip <- function(data,N.sample,  ini.modCount=NULL, ini.modZero=NULL,
                  dist="poisson", link="logit", neighb, prob, cc, ...)
  {
## ---------------------------------------------------------------
## Arguments: data: a data.frame or matrix which is used in model
##                  selection. The first column is the response variable. 
##                  The other columns and a column of
##                  1's (not included in "data") comprise the design matrix
##         	    for the full model.
##  N.sample: sample size of the Metropolis-Hasting sample.
##  ini.modCount    initial model for the count part
##  ini.modZero     initial model for the presence/absence part
##  dist    distribution for the count part of the model in function pscl:zeroinfl
##  link    link function for the count part of the model in function pscl:zeroinfl
##  neighb: type of neighborhood for the MCMC chain. Available choices
##          are "dropadd1" and "all".
##  prob: probabilities scheme for the sampling of the next model in the MCMC
##          chain. Available choices are: "equal", "pvalue".
##  cc      tuning constant value      
## ---------------------------------------------------------------

    subset <- function(k)
      {
        m <- matrix(c(0, 1), nrow = 2)
        for(i in 1:(k - 1))
          {
            m <- rbind(cbind(0, m), cbind(1, m))
          }
        m1 <- m
        p <- apply(m1, 1, sum) + 1
        s <- m1[order(p),  ]
        s <- s[ - (p == 1),  ]
        mode(s) <- "logical"
        rbind(rep(F,k),s)
      }    

# PRELIMINARY TESTS
    neighbs<-c("dropadd1", "all")
    if(as.integer(match(c(neighb), neighbs, -1))<1)
      stop("unknown neighborhood")
    probs <- c("equal", "pvalue")
    if(as.integer(match(c(prob), probs, -1))<1)
      stop("unknown probability scheme")

	ncolx <- ncol(data) - 1
	dbcol <- (2*ncolx)
	nrowx <- nrow(data)
	zip <- summary(zeroinfl(data[,1]~as.matrix(data[,-1]),dist=dist,link=link))
        if(dist=="poisson")
          {
            pv <- as.vector(cbind(zip$coeff$count[-1,4],zip$coeff$zero[-1,4]))
          }
        if(dist=="negbin")
          {
            pv <- as.vector(cbind(zip$coeff$count[-c(1,ncolx+2),4],zip$coeff$zero[-1,4]))
          }

    if(is.null(ini.modCount))
      ini.modCount <- rep(1, ncolx)
    if(is.null(ini.modZero))
	ini.modZero <- rep(1,ncolx)
    model.sample <- matrix(-1, N.sample, (dbcol + 2))
	
	for(i in 1:N.sample)
        {
            if(i == 1)
            {
                modC <- ini.modCount
		modZ <- ini.modZero
		mod <- as.vector(cbind(modC,modZ))
		y = data[,1]
		x1 = as.matrix(cbind(rep(1, nrowx), data[,  - 1][modC==1]))
		x2 = as.matrix(cbind(rep(1, nrowx), data[,  - 1][modZ==1]))
		ini.aic <- AIC(myzeroinfl(y~-1+x1|-1+x2,dist=dist,link=link), k = sum(mod))
		aic <- ini.aic
		ini.mod <- mod
            }
            else
                {
                    mod <- model.sample[(i - 1),  - ((dbcol + 1):(dbcol + 2))]
                    aic <- model.sample[(i  - 1), (dbcol + 1)]
		}
        if(neighb=="dropadd1")
          {
            neighbmodels <- matrix(rep(mod,dbcol),nrow=dbcol,ncol=dbcol,byrow=T)
            diag(neighbmodels) <- !diag(neighbmodels)
          }
        else
          neighbmodels <- models

        if(prob=="equal")
          probab <- rep(1,nrow(neighbmodels))
        else
          {
            ind.add <- apply(neighbmodels,1,function(thisrow,model){sum(thisrow)>sum(model)},model=mod)
            probab <- (1-pv)*ind.add + pv*(!ind.add)
            probab <- probab/sum(probab)
          }

        ind.mod <- sample(nrow(neighbmodels),1,prob=probab)
        newmod <- neighbmodels[ind.mod,]
        newmodC <- newmod[1:ncolx]
        newmodZ <- newmod[(ncolx+1):dbcol]
        x1 = as.matrix(cbind(rep(1, nrowx), data[, - 1][newmodC==1]))
        x2 = as.matrix(cbind(rep(1, nrowx), data[, - 1][newmodZ==1]))
        newaic <- AIC(myzeroinfl(y~-1+x1|-1+x2,dist=dist,link=link), k = sum(newmod))
        acc <- 0
        unif <- runif(1)
        if(prob=="equal")
          cutoff <- min(exp(cc*(aic-newaic)),1)
        else
          cutoff <- min(exp(cc*(aic-newaic))*(1-probab[ind.mod])/probab[ind.mod],1)
        if(unif<cutoff)
          {
            acc <- 1 # For the acceptance rate
            mod <- newmod
            aic <- newaic
          }
 	model.sample[i, ] <- c(mod, aic, acc)
	}
	model.sample <- rbind(c(ini.mod,ini.aic,0),model.sample)
	dimnames(model.sample) <- list(NULL, c(paste("x", 1:dbcol, sep = ""), 
		"AIC","acc"))
	return(model.sample)
}

# zipMH function
# Funtion to generate the Markov chain, when using CV
zipMH <- function(data, M.splits=50, N.sample, ini.modCount=NULL, dist="poisson", ini.modZero=NULL,neighb, prob, cc, ...)
  {
## ---------------------------------------------------------------
## Arguments: data: a data.frame or matrix which is used in model
##                  selection. The first column is the response
##                  variable.The other columns and a column of
##                  1's (not included in "data") comprise the design matrix
##         	    for the full model.
##  M.splits: number of splits to be used for cross-validation procedure.
##  N.sample: sample size of the Metropolis-Hasting sample.
##  models: a matrix (2^p x p) of T and F with all the possible submodels.
##  ini.modCount    initial model for the count part
##  ini.modZero     initial model for the presence/absence part
##  dist    distribution for the count part of the model in function pscl:zeroinfl
##  link    link function for the count part of the model in function pscl:zeroinfl
##  neighb: type of neighborhood for the MCMC chain. Available choices
##          are "dropadd1" and "all".
##  prob: probabilities scheme for the sampling of the next model in the MCMC
##          chain. Available choices are: "equal", "pvalue".
##  cc      tuning constant value      
## ---------------------------------------------------------------

# AUXILIARY FUNCTIONS
gensplits <- function(data,M)
     {
       # This function computes the M splits of the data required for the
       # computation of the prediction error via cross validation.
       # Note:	 id must be a vector of integers in sequence from 1 through K
       K <- nrow(data) 
       Kc <- round(K/2,0)  
       splitmat <- matrix(NA,nrow=K,ncol=M)
	 index <- c(1:nrow(data))
       for(i in 1:M)
         {
           thissubset <- sample(K,Kc,replace=F)
           splitmat[,i] <-  index %in% thissubset
         }
       return(splitmat)
     }

pe.f <- function(y,z,x,M,splits,dist="poisson",link="logit") 
      {
        # This function is to compute the prediction error criterion value of a
        # zip model.
        pec <- c(rep(0,M))
        for(i in 1:M)
         
              {
                  thisfit <- myzeroinfl(y~-1+z|-1+x,dist=dist,link=link,subset=splits[,i])
                  mu_hat <- fitted(thisfit,type="response")

			thisY <- y[!splits[,i]]
			thisfitted <- mu_hat 
			thispe <- mean((thisY - thisfitted)^2)
            	pec[i] <- thispe
          }
        avpe <- mean(pec)
	    sdpe <- sqrt(var(pec)/M)
        return(c(avpe,sdpe))
      }
    
# PRELIMINARY TESTS
    neighbs<-c("dropadd1", "all")
    if(as.integer(match(c(neighb), neighbs, -1))<1)
      stop("unknown neighborhood")
    probs<-c("equal", "pvalue")
    if(as.integer(match(c(prob), probs, -1))<1)
      stop("unknown probability scheme")
     

# MAIN ROUTINE:
     mysplit <- gensplits(data, M=M.splits)
     ncolx <- ncol(data) - 1
     dbcol <- (2*ncolx)
     nrowx <- nrow(data)

	zip <- summary(zeroinfl(data[,1]~as.matrix(data[,-1]),dist=dist,link="logit"))
        if(dist=="poisson")
          {
            pv <- as.vector(cbind(zip$coeff$count[-1,4],zip$coeff$zero[-1,4]))
          }
        if(dist=="negbin")
          {
            pv <- as.vector(cbind(zip$coeff$count[-c(1,ncolx+2),4],zip$coeff$zero[-1,4]))
          }
   
    if(is.null(ini.modCount))
      ini.modCount <- rep(1, ncolx)
    if(is.null(ini.modZero))
	ini.modZero <- rep(1,ncolx)
    model.sample <- matrix(-1, (N.sample), (dbcol + 3))

    for(i in 1:(N.sample))
      {
	  flush.console()
        if(i == 1)
          {
            modC <- ini.modCount
	    modZ <- ini.modZero
            pe <- pe.f(y = data[, 1],z = as.matrix(cbind(rep(1, nrowx), data[, - 1][modC==1])),
		x = as.matrix(cbind(rep(1, nrowx), data[,  - 1][modZ==1])),splits=mysplit, dist=dist, M=M.splits)
            ini.Mpe <- pe[1]
	    ini.Spe <- pe[2]
	    Mpe <- ini.Mpe
     	    Spe <- ini.Spe
	    mod=as.vector(cbind(modC,modZ))  
          }
        else
          {
            mod <- model.sample[(i - 1),  - ((dbcol + 1):(dbcol + 3))]
            Mpe <- model.sample[(i  - 1), (dbcol+ 1)]
	    Spe <- model.sample[(i  - 1), (dbcol+ 2)]
          }

        if(neighb=="dropadd1")
          {
            neighbmodels <- matrix(rep(mod,dbcol),nrow=dbcol,ncol=dbcol,byrow=T)
            diag(neighbmodels) <- !diag(neighbmodels)
          }
        else
          neighbmodels <- models

        if(prob=="equal")
          probab <- rep(1,nrow(neighbmodels))
        else
          {
            ind.add <- apply(neighbmodels,1,function(thisrow,model){sum(thisrow)>sum(model)},model=mod)
            probab <- (1-pv)*ind.add + pv*(!ind.add)
            probab <- probab/sum(probab)
          }

        ind.mod <- sample(nrow(neighbmodels),1,prob=probab)
        newmod <- neighbmodels[ind.mod,]
        newmodC <- newmod[1:ncolx]
        newmodZ <- newmod[(ncolx+1):dbcol]

        newpe <- pe.f(y = data[, 1], z = as.matrix(cbind(rep(1, nrowx), data[,  - 1][newmodC==1])),
        x = as.matrix(cbind(rep(1, nrowx), data[,  - 1][newmodZ==1])),splits=mysplit, dist=dist, M=M.splits)
        newMpe <- newpe[1]
        newSpe <- newpe[2]
        acc <-0
        unif <- runif(1)
        if(prob=="equal")
          cutoff <- min(exp(cc*(Mpe-newMpe)/ini.Spe),1)
        else
	    cutoff <- min(exp(cc*(Mpe-newMpe)/ini.Spe)*(1-probab[ind.mod])/probab[ind.mod],1)
        if(unif<cutoff)
          {
		    acc <- 1 # For the acceptance rate
            mod <- newmod
            Mpe <- newMpe
		    Spe <- newSpe
          }
	  flush.console()
        model.sample[i, ] <- c(mod, Mpe, Spe, acc)
      }

    model.sample <- rbind(c(ini.modCount,ini.modZero,ini.Mpe, ini.Spe, 0),model.sample)
    dimnames(model.sample) <- list(NULL, c(paste("x", 1:dbcol, sep = ""),"AvPE","SdPE","acc"))
    return(model.sample)
  }

# myzeroinfl() is a modified version of pscl:zeroinfl
#####################################################
myzeroinfl <- function (formula, data, subset, na.action, weights, offset, 
    dist = c("poisson", "negbin", "geometric"), link = c("logit", 
        "probit", "cloglog", "cauchit", "log"), control = zeroinfl.control(...), 
    model = TRUE, y = TRUE, x = FALSE, ...) 
{
    ziPoisson <- function(parms) {
        mu <- as.vector(exp(X %*% parms[1:kx] + offsetx))
        phi <- as.vector(linkinv(Z %*% parms[(kx + 1):(kx + kz)] + 
            offsetz))
        loglik0 <- log(phi + exp(log(1 - phi) - mu))
        loglik1 <- log(1 - phi) + dpois(Y, lambda = mu, log = TRUE)
        loglik <- sum(weights[Y0] * loglik0[Y0]) + sum(weights[Y1] * 
            loglik1[Y1])
        loglik
    }
    ziNegBin <- function(parms) {
        mu <- as.vector(exp(X %*% parms[1:kx] + offsetx))
        phi <- as.vector(linkinv(Z %*% parms[(kx + 1):(kx + kz)] + 
            offsetz))
        theta <- exp(parms[(kx + kz) + 1])
        loglik0 <- log(phi + exp(log(1 - phi) + suppressWarnings(dnbinom(0, 
            size = theta, mu = mu, log = TRUE))))
        loglik1 <- log(1 - phi) + suppressWarnings(dnbinom(Y, 
            size = theta, mu = mu, log = TRUE))
        loglik <- sum(weights[Y0] * loglik0[Y0]) + sum(weights[Y1] * 
            loglik1[Y1])
        loglik
    }
    ziGeom <- function(parms) ziNegBin(c(parms, 0))
    gradPoisson <- function(parms) {
        eta <- as.vector(X %*% parms[1:kx] + offsetx)
        mu <- exp(eta)
        etaz <- as.vector(Z %*% parms[(kx + 1):(kx + kz)] + offsetz)
        muz <- linkinv(etaz)
        clogdens0 <- -mu
        dens0 <- muz * (1 - as.numeric(Y1)) + exp(log(1 - muz) + 
            clogdens0)
        wres_count <- ifelse(Y1, Y - mu, -exp(-log(dens0) + log(1 - 
            muz) + clogdens0 + log(mu)))
        wres_zero <- ifelse(Y1, -1/(1 - muz) * linkobj$mu.eta(etaz), 
            (linkobj$mu.eta(etaz) - exp(clogdens0) * linkobj$mu.eta(etaz))/dens0)
        colSums(cbind(wres_count * weights * X, wres_zero * weights * 
            Z))
    }
    gradGeom <- function(parms) {
        eta <- as.vector(X %*% parms[1:kx] + offsetx)
        mu <- exp(eta)
        etaz <- as.vector(Z %*% parms[(kx + 1):(kx + kz)] + offsetz)
        muz <- linkinv(etaz)
        clogdens0 <- dnbinom(0, size = 1, mu = mu, log = TRUE)
        dens0 <- muz * (1 - as.numeric(Y1)) + exp(log(1 - muz) + 
            clogdens0)
        wres_count <- ifelse(Y1, Y - mu * (Y + 1)/(mu + 1), -exp(-log(dens0) + 
            log(1 - muz) + clogdens0 - log(mu + 1) + log(mu)))
        wres_zero <- ifelse(Y1, -1/(1 - muz) * linkobj$mu.eta(etaz), 
            (linkobj$mu.eta(etaz) - exp(clogdens0) * linkobj$mu.eta(etaz))/dens0)
        colSums(cbind(wres_count * weights * X, wres_zero * weights * 
            Z))
    }
    gradNegBin <- function(parms) {
        eta <- as.vector(X %*% parms[1:kx] + offsetx)
        mu <- exp(eta)
        etaz <- as.vector(Z %*% parms[(kx + 1):(kx + kz)] + offsetz)
        muz <- linkinv(etaz)
        theta <- exp(parms[(kx + kz) + 1])
        clogdens0 <- dnbinom(0, size = theta, mu = mu, log = TRUE)
        dens0 <- muz * (1 - as.numeric(Y1)) + exp(log(1 - muz) + 
            clogdens0)
        wres_count <- ifelse(Y1, Y - mu * (Y + theta)/(mu + theta), 
            -exp(-log(dens0) + log(1 - muz) + clogdens0 + log(theta) - 
                log(mu + theta) + log(mu)))
        wres_zero <- ifelse(Y1, -1/(1 - muz) * linkobj$mu.eta(etaz), 
            (linkobj$mu.eta(etaz) - exp(clogdens0) * linkobj$mu.eta(etaz))/dens0)
        wres_theta <- theta * ifelse(Y1, digamma(Y + theta) - 
            digamma(theta) + log(theta) - log(mu + theta) + 1 - 
            (Y + theta)/(mu + theta), exp(-log(dens0) + log(1 - 
            muz) + clogdens0) * (log(theta) - log(mu + theta) + 
            1 - theta/(mu + theta)))
        colSums(cbind(wres_count * weights * X, wres_zero * weights * 
            Z, wres_theta))
    }
    dist <- match.arg(dist)
    loglikfun <- switch(dist, poisson = ziPoisson, geometric = ziGeom, 
        negbin = ziNegBin)
    gradfun <- switch(dist, poisson = gradPoisson, geometric = gradGeom, 
        negbin = gradNegBin)
    linkstr <- match.arg(link)
    linkobj <- make.link(linkstr)
    linkinv <- linkobj$linkinv
    if (control$trace) 
        cat("Zero-inflated Count Model\n", paste("count model:", 
            dist, "with log link\n"), paste("zero-inflation model: binomial with", 
            linkstr, "link\n"), sep = "")
    cl <- match.call()
    if (missing(data)) 
        data <- environment(formula)
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action", "weights", 
        "offset"), names(mf), 0)
    mf <- mf[c(1, m)]
    mf$drop.unused.levels <- TRUE
    if (length(formula[[3]]) > 1 && identical(formula[[3]][[1]], 
        as.name("|"))) {
        ff <- formula
        formula[[3]][1] <- call("+")
        mf$formula <- formula
        ffc <- . ~ .
        ffz <- ~.
        ffc[[2]] <- ff[[2]]
        ffc[[3]] <- ff[[3]][[2]]
        ffz[[3]] <- ff[[3]][[3]]
        ffz[[2]] <- NULL
    }
    else {
        ffz <- ffc <- ff <- formula
        ffz[[2]] <- NULL
    }
    if (inherits(try(terms(ffz), silent = TRUE), "try-error")) {
        ffz <- eval(parse(text = sprintf(paste("%s -", deparse(ffc[[2]])), 
            deparse(ffz))))
    }
    mf[[1]] <- as.name("model.frame")
    mf <- eval(mf, parent.frame())
    mt <- attr(mf, "terms")
    mtX <- terms(ffc, data = data)
    X <- model.matrix(mtX, mf)
    mtZ <- terms(ffz, data = data)
    mtZ <- terms(update(mtZ, ~.), data = data)
    Z <- model.matrix(mtZ, mf)
    Y <- model.response(mf, "numeric")
    if (length(Y) < 1) 
        stop("empty model")
    if (all(Y > 0)) 
        stop("invalid dependent variable, minimum count is not zero")
    if (!isTRUE(all.equal(as.vector(Y), as.integer(round(Y + 
        0.001))))) 
        stop("invalid dependent variable, non-integer values")
    Y <- as.integer(round(Y + 0.001))
    if (any(Y < 0)) 
        stop("invalid dependent variable, negative counts")
    if (control$trace) {
        cat("dependent variable:\n")
        tab <- table(factor(Y, levels = 0:max(Y)), exclude = NULL)
        names(dimnames(tab)) <- NULL
        print(tab)
    }
    n <- length(Y)
    kx <- NCOL(X)
    kz <- NCOL(Z)
    Y0 <- Y <= 0
    Y1 <- Y > 0
    weights <- model.weights(mf)
    if (is.null(weights)) 
        weights <- 1
    if (length(weights) == 1) 
        weights <- rep.int(weights, n)
    weights <- as.vector(weights)
    names(weights) <- rownames(mf)
    offsetx <- pscl:::model_offset_2(mf, terms = mtX, offset = TRUE)
    if (is.null(offsetx)) 
        offsetx <- 0
    if (length(offsetx) == 1) 
        offsetx <- rep.int(offsetx, n)
    offsetx <- as.vector(offsetx)
    offsetz <- pscl:::model_offset_2(mf, terms = mtZ, offset = FALSE)
    if (is.null(offsetz)) 
        offsetz <- 0
    if (length(offsetz) == 1) 
        offsetz <- rep.int(offsetz, n)
    offsetz <- as.vector(offsetz)
    start <- control$start
    if (!is.null(start)) {
        valid <- TRUE
        if (!("count" %in% names(start))) {
            valid <- FALSE
            warning("invalid starting values, count model coefficients not specified")
            start$count <- rep.int(0, kx)
        }
        if (!("zero" %in% names(start))) {
            valid <- FALSE
            warning("invalid starting values, zero-inflation model coefficients not specified")
            start$zero <- rep.int(0, kz)
        }
        if (length(start$count) != kx) {
            valid <- FALSE
            warning("invalid starting values, wrong number of count model coefficients")
        }
        if (length(start$zero) != kz) {
            valid <- FALSE
            warning("invalid starting values, wrong number of zero-inflation model coefficients")
        }
        if (dist == "negbin") {
            if (!("theta" %in% names(start))) 
                start$theta <- 1
            start <- list(count = start$count, zero = start$zero, 
                theta = as.vector(start$theta[1]))
        }
        else {
            start <- list(count = start$count, zero = start$zero)
        }
        if (!valid) 
            start <- NULL
    }
    if (is.null(start)) {
        if (control$trace) 
            cat("generating starting values...")
        model_count <- glm.fit(X, Y, family = poisson(), weights = weights, 
            offset = offsetx)
        model_zero <- glm.fit(Z, as.integer(Y0), weights = weights, 
            family = binomial(link = linkstr), offset = offsetz)
        start <- list(count = model_count$coefficients, zero = model_zero$coefficients)
        if (dist == "negbin") 
            start$theta <- 1
        if (control$EM & dist == "poisson") {
            mui <- model_count$fitted
            probi <- model_zero$fitted
            probi <- probi/(probi + (1 - probi) * dpois(0, mui))
            probi[Y1] <- 0
            ll_new <- loglikfun(c(start$count, start$zero))
            ll_old <- 2 * ll_new
            while (abs((ll_old - ll_new)/ll_old) > control$reltol) {
                ll_old <- ll_new
                model_count <- glm.fit(X, Y, weights = weights * 
                  (1 - probi), offset = offsetx, family = poisson(), 
                  start = start$count)
                model_zero <- suppressWarnings(glm.fit(Z, probi, 
                  weights = weights, offset = offsetz, family = binomial(link = linkstr), 
                  start = start$zero))
                mui <- model_count$fitted
                probi <- model_zero$fitted
                probi <- probi/(probi + (1 - probi) * dpois(0, 
                  mui))
                probi[Y1] <- 0
                start <- list(count = model_count$coefficients, 
                  zero = model_zero$coefficients)
                ll_new <- loglikfun(c(start$count, start$zero))
            }
        }
        if (control$EM & dist == "geometric") {
            mui <- model_count$fitted
            probi <- model_zero$fitted
            probi <- probi/(probi + (1 - probi) * dnbinom(0, 
                size = 1, mu = mui))
            probi[Y1] <- 0
            ll_new <- loglikfun(c(start$count, start$zero))
            ll_old <- 2 * ll_new
            if (!require("MASS")) {
                ll_old <- ll_new
                warning("EM estimation of starting values not available")
            }
            while (abs((ll_old - ll_new)/ll_old) > control$reltol) {
                ll_old <- ll_new
                model_count <- suppressWarnings(glm.fit(X, Y, 
                  weights = weights * (1 - probi), offset = offsetx, 
                  family = negative.binomial(1), start = start$count))
                model_zero <- suppressWarnings(glm.fit(Z, probi, 
                  weights = weights, offset = offsetz, family = binomial(link = linkstr), 
                  start = start$zero))
                start <- list(count = model_count$coefficients, 
                  zero = model_zero$coefficients)
                mui <- model_count$fitted
                probi <- model_zero$fitted
                probi <- probi/(probi + (1 - probi) * dnbinom(0, 
                  size = 1, mu = mui))
                probi[Y1] <- 0
                ll_new <- loglikfun(c(start$count, start$zero))
            }
        }
        if (control$EM & dist == "negbin") {
            mui <- model_count$fitted
            probi <- model_zero$fitted
            probi <- probi/(probi + (1 - probi) * dnbinom(0, 
                size = start$theta, mu = mui))
            probi[Y1] <- 0
            ll_new <- loglikfun(c(start$count, start$zero, log(start$theta)))
            ll_old <- 2 * ll_new
            if (!require("MASS")) {
                ll_old <- ll_new
                warning("EM estimation of starting values not available")
            }
            offset <- offsetx
            while (abs((ll_old - ll_new)/ll_old) > control$reltol) {
                ll_old <- ll_new
                model_count <- suppressWarnings(glm.nb(Y ~ 0 + 
                  X + offset(offset), weights = weights * (1 - 
                  probi), start = start$count, init.theta = start$theta))
                model_zero <- suppressWarnings(glm.fit(Z, probi, 
                  weights = weights, offset = offsetz, family = binomial(link = linkstr), 
                  start = start$zero))
                start <- list(count = model_count$coefficients, 
                  zero = model_zero$coefficients, theta = model_count$theta)
                mui <- model_count$fitted
                probi <- model_zero$fitted
                probi <- probi/(probi + (1 - probi) * dnbinom(0, 
                  size = start$theta, mu = mui))
                probi[Y1] <- 0
                ll_new <- loglikfun(c(start$count, start$zero, 
                  log(start$theta)))
            }
        }
        if (control$trace) 
            cat("done\n")
    }
    if (control$trace) 
        cat("calling optim() for ML estimation:\n")
    method <- control$method
    hessian <- control$hessian
    ocontrol <- control
    control$method <- control$hessian <- control$EM <- control$start <- NULL
    fit <- optim(fn = loglikfun, gr = gradfun, par = c(start$count, 
        start$zero, if (dist == "negbin") log(start$theta) else NULL), 
        method = method, hessian = hessian, control = control)
    if (fit$convergence > 0) 
        warning("optimization failed to converge")
    coefc <- fit$par[1:kx]
    names(coefc) <- names(start$count) <- colnames(X)
    coefz <- fit$par[(kx + 1):(kx + kz)]
    names(coefz) <- names(start$zero) <- colnames(Z)
#    vc <- -solve(as.matrix(fit$hessian))
    if (dist == "negbin") {
        np <- kx + kz + 1
        theta <- as.vector(exp(fit$par[np]))
        SE.logtheta <- NULL
#        SE.logtheta <- as.vector(sqrt(diag(vc)[np]))
#        vc <- vc[-np, -np, drop = FALSE]
    }
    else {
        theta <- NULL
       SE.logtheta <- NULL
    }
#    colnames(vc) <- rownames(vc) <- c(paste("count", colnames(X), 
#        sep = "_"), paste("zero", colnames(Z), sep = "_"))
    mu <- exp(X %*% coefc + offsetx)[, 1]
    phi <- linkinv(Z %*% coefz + offsetz)[, 1]
    Yhat <- (1 - phi) * mu
    res <- sqrt(weights) * (Y - Yhat)
    nobs <- sum(weights > 0)
    rval <- list(coefficients = list(count = coefc, zero = coefz), 
        residuals = res, fitted.values = Yhat, optim = fit, method = method, 
        control = ocontrol, start = start, weights = if (identical(as.vector(weights), 
            rep.int(1L, n))) NULL else weights, offset = list(count = if (identical(offsetx, 
            rep.int(0, n))) NULL else offsetx, zero = if (identical(offsetz, 
            rep.int(0, n))) NULL else offsetz), n = nobs, df.null = nobs - 
            2, df.residual = nobs - (kx + kz + (dist == "negbin")), 
        terms = list(count = mtX, zero = mtZ, full = mt), theta = theta, 
        SE.logtheta = SE.logtheta, loglik = fit$value, #vcov = vc, 
        dist = dist, link = linkstr, linkinv = linkinv, converged = fit$convergence < 
            1, call = cl, formula = ff, levels = .getXlevels(mt, 
            mf), contrasts = list(count = attr(X, "contrasts"), 
            zero = attr(Z, "contrasts")))
    if (model) 
        rval$model <- mf
    if (y) 
        rval$y <- Y
    if (x) 
        rval$x <- list(count = X, zero = Z)
    class(rval) <- "zeroinfl"
    return(rval)
}




# Function that allows to compute the votes
# from https://stat.ethz.ch/pipermail/r-help/2008-January/151489.html
# (contribution of Charles C. Berry)
############################################
count.rows <- function(x)
   {
     order.x <- do.call(order,as.data.frame(x))
     equal.to.previous <-
       rowSums(x[tail(order.x,-1),] != x[head(order.x,-1),])==0
     tf.runs <- rle(equal.to.previous)
     counts <- c(1,
                 unlist(mapply( function(x,y) if (y) x+1 else (rep(1,x)),
                               tf.runs$length, tf.runs$value )))
     counts <- counts[ c(diff(counts) <= 0, TRUE ) ]
     unique.rows <- which( c(TRUE, !equal.to.previous ) )
     cbind( counts, x[order.x[ unique.rows ], ,drop=F] )
   }

