##################################################################################
# ECR - January 2017
#
# This is the code accompanying
# E. Cantoni & M. Auda (2017)
# "Stochastic Variable Selection selection strategies for zero-inflated models"
#################################################################################

# Load data from the zic package
require("zic")
data(docvisits)

# define the model and the associated dataframe built with model.matrix().
# For example:
modelNB.formula <- docvisits ~ age + health + handicap + hdegree 

# or a more realistic, but longer to fit:
# modelNB.formula <- docvisits ~ age + agesq + health + handicap + hdegree + married + schooling + hhincome + children + self + civil + bluec + employed + public + addon 
docs.NB <- as.data.frame(cbind(docvisits$docvisits,model.matrix(modelNB.formula, data = docvisits)[,-1]))
dimnames(docs.NB)[[2]][1] <- "docvisits"

# Load pscl package to use zeroinfl() function
require("pscl")

# Source R functions
source("StocVarSelFcts.R")
# see this file for a description of the function arguments

# Variable selection with AIC
##############################
chainlength <- 5000
doczipAIC.NB <- aicMHzip(docs.NB,N.sample=chainlength, ini.modCount=NULL, ini.modZero=NULL,neighb="dropadd1", dist="negbin", prob="pvalue", cc=0.25)
dimnames(doczipAIC.NB)[[2]][1:((ncol(docs.NB)*2)-2)] <- c(dimnames(docs.NB)[[2]][-1],dimnames(docs.NB)[[2]][-1])


# Analyzing and summarizing results for AIC
###########################################

# Marginal frequencies for AIC on the entire chain
margfreqAIC.NB <- apply(doczipAIC.NB[,1:((ncol(docs.NB)*2)-2)],2,mean)
margfreqAIC.NB

# 50% consensus model for AIC
consensus50AIC.NB <- (margfreqAIC.NB>0.5)
consensus50AIC.NB
# first part corresponds to count equation, second part to binary equation

# Minimum criterion for AIC
whichminAIC.NB <- min(c(which(doczipAIC.NB[,"AIC"]==min(doczipAIC.NB[,"AIC"]))))
minAIC.NB <- doczipAIC.NB[whichminAIC.NB,]
minAIC.NB

# Model votes for AIC
# Burn-in period neeed to be removed
burnin <- 1000 
margfreAICwo.NB <- apply(doczipAIC.NB[(burnin+1):(chainlength+1),1:((ncol(docs.NB)*2)-2)],2,mean)
margfreAICwo.NB


resAIC.NB <- count.rows(doczipAIC.NB[(burnin+1):(chainlength+1),-((ncol(docs.NB)*2))])
dAIC.NB <- which(resAIC.NB[,1]==max(resAIC.NB[,1]))
lAIC.NB <- length(dAIC.NB)
uniqueAIC.NB <- nrow(resAIC.NB)
minAIC.NB <- min(resAIC.NB[,"AIC"])
if(lAIC.NB==1)
  {
    votesAIC.NB <- list(model=resAIC.NB[dAIC.NB,],AIC=minAIC.NB,uniqueAIC.NB)
    } else
  {
   votesAIC.NB <- list(model=resAIC.NB[dAIC.NB,],AIC=rep(minAIC.NB,lAIC.NB),rep(uniqueAIC.NB,lAIC.NB))
  }
votesAIC.NB$model


# Variable selection with CV
#############################
M <- 20 
doczipCV.NB <- zipMH(docs.NB, M.splits=M, N.sample=chainlength, ini.modCount=NULL, ini.modZero=NULL,neighb="dropadd1", dist="negbin", prob="pvalue", cc=20)
# depending on the value of M and the size of the problem, this can be long
dimnames(doczipCV.NB)[[2]][1:((ncol(docs.NB)*2)-2)] <- c(dimnames(docs.NB)[[2]][-1],dimnames(docs.NB)[[2]][-1])

# Analyzing and summarizing results for CV
##########################################

# Marginal frequencies for CV
margfreqCV.NB <- apply(doczipCV.NB[,1:((ncol(docs.NB)*2)-2)],2,mean)
margfreqCV.NB

# 50% consensus model for CV
consensus50CV.NB <- (margfreqCV.NB>0.5)
consensus50CV.NB
# first part corresponds to count equation, second part to binary equation

# Minimum criterion for CV
whichminCV.NB <- min(c(which(doczipCV.NB[,"AvPE"]==min(doczipCV.NB[,"AvPE"]))))
minCV.NB <- doczipCV.NB[whichminCV.NB,]

minCV.NB

# Model votes for CV
# Burn-in period neeed to be removed
burnin <- 1000 
margfreqCVwo.NB <- apply(doczipAIC.NB[(burnin+1):(chainlength+1),1:((ncol(docs.NB)*2)-2)],2,mean)
margfreAICwo.NB

resCV.NB <- count.rows(doczipCV.NB[(burnin+1):(chainlength+1),-((ncol(docs.NB)*2))])
dCV.NB <- which(resCV.NB[,1]==max(resCV.NB[,1]))
lCV.NB <- length(dCV.NB)
uniqueCV.NB <- nrow(resCV.NB)
minCV.NB <- min(resCV.NB[,"AvPE"])
if(lCV.NB==1)
  {
    votesCV.NB <- list(model=resCV.NB[dCV.NB,],PE=minCV.NB,uniqueCV.NB)
    } else
  {
   votesCV.NB <- list(model=resCV.NB[dCV.NB,],PE=rep(minCV.NB,lCV.NB),rep(uniqueCV.NB,lCV.NB))
  }
votesCV.NB$model


# Analysis with the mpath package
#######################################
require(mpath)
docLASSO.Wang <- zipath(docvisits ~ . | ., data = docs.NB, family = "negbin", nlambda = 100, lambda.zero.min.ratio = 0.001, maxit.em = 300, 
maxit.theta = 25, theta.fixed = FALSE, trace=FALSE, penalty="enet", rescale=FALSE)
minBIC.docLASSO.Wang <- which.min(BIC(docLASSO.Wang))
docLASSO.Wang.coeff <- c(coef(docLASSO.Wang,minBIC.docLASSO.Wang)$count[-1],coef(docLASSO.Wang,minBIC.docLASSO.Wang)$zero[-1])


docMCP.Wang <- zipath(docvisits ~ . | ., data = docs.NB, family = "negbin", nlambda = 100, lambda.zero.min.ratio = 0.001, maxit.em = 300,  
maxit.theta = 25, theta.fixed = FALSE, trace=FALSE, penalty="mnet", rescale=FALSE)
minBIC.docMCP.Wang <- which.min(BIC(docMCP.Wang))
docMCP.Wang.coeff <-c(coef(docMCP.Wang,minBIC.docMCP.Wang)$count[-1],coef(docMCP.Wang,minBIC.docMCP.Wang)$zero[-1])


docSCAD.Wang <- zipath(docvisits ~ . | ., data = docs.NB, family = "negbin", nlambda = 100, lambda.zero.min.ratio = 0.001, maxit.em = 300,  
maxit.theta = 25, theta.fixed = FALSE, trace=FALSE, penalty="snet", rescale=FALSE)
minBIC.docSCAD.Wang <- which.min(BIC(docSCAD.Wang))
docSCAD.Wang.coeff <-c(coef(docSCAD.Wang,minBIC.docSCAD.Wang)$count[-1],coef(docSCAD.Wang,minBIC.docSCAD.Wang)$zero[-1])





